package eu.mhsl.craftattack.spawn;

import eu.mhsl.craftattack.spawn.api.HttpServer;
import eu.mhsl.craftattack.spawn.appliance.Appliance;
import eu.mhsl.craftattack.spawn.config.Configuration;
import org.bukkit.Bukkit;
import org.bukkit.event.HandlerList;
import org.bukkit.plugin.java.JavaPlugin;
import org.reflections.Reflections;

import java.lang.reflect.ParameterizedType;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

public final class Main extends JavaPlugin {
    private static Main instance;
    private static Logger logger;

    private List<Appliance> appliances;
    private HttpServer httpApi;

    @Override
    public void onEnable() {
        instance = this;
        logger = instance().getLogger();
        saveDefaultConfig();
        try {
            this.wrappedEnable();
        } catch (Exception e) {
            Main.logger().log(Level.SEVERE, "Error while initializing Spawn plugin, shutting down!", e);
            Bukkit.shutdown();
        }
    }

    private void wrappedEnable() {
        Configuration.readConfig();
        List<String> disabledAppliances = Configuration.pluginConfig.getStringList("disabledAppliances");

        Main.logger.info("Loading appliances...");
        Reflections reflections = new Reflections(this.getClass().getPackageName());
        Set<Class<? extends Appliance>> applianceClasses = reflections.getSubTypesOf(Appliance.class);

        this.appliances = applianceClasses.stream()
            .filter(applianceClass -> !disabledAppliances.contains(applianceClass.getSimpleName()))
            .map(applianceClass -> {
                try {
                    return (Appliance) applianceClass.getDeclaredConstructor().newInstance();
                } catch (Exception e) {
                    throw new RuntimeException(String.format("Failed to create instance of '%s'", applianceClass.getName()), e);
                }
            })
            .toList();
        Main.logger().info(String.format("Loaded %d appliances!", appliances.size()));

        Main.logger().info("Initializing appliances...");
        this.appliances.forEach(appliance -> {
            appliance.onEnable();
            appliance.initialize(this);
        });
        Main.logger().info(String.format("Initialized %d appliances!", appliances.size()));

        Main.logger().info("Starting HTTP API...");
        this.httpApi = new HttpServer();

        getServer().getMessenger().registerOutgoingPluginChannel(this, "BungeeCord");
        Main.logger().info("Startup complete!");
    }

    @Override
    public void onDisable() {
        Main.logger().info("Disabling appliances...");
        appliances.forEach(appliance -> {
            Main.logger().info("Disabling " + appliance.getClass().getSimpleName());
            appliance.onDisable();
            appliance.destruct(this);
        });
        this.httpApi.stop();
        HandlerList.unregisterAll(this);
        Bukkit.getScheduler().cancelTasks(this);
        Main.logger().info("Disabled " + appliances.size() + " appliances!");
    }

    public <T extends Appliance> T getAppliance(Class<T> clazz) {
        return this.appliances.stream()
            .filter(clazz::isInstance)
            .map(clazz::cast)
            .findFirst()
            .orElseThrow(() -> new RuntimeException(String.format("Appliance %s not loaded or instantiated!", clazz)));
    }

    @SuppressWarnings("unchecked")
    public static <T> Class<T> getApplianceType(Class<?> clazz) {
        return (Class<T>) ((ParameterizedType) clazz.getGenericSuperclass()).getActualTypeArguments()[0];
    }

    public List<Appliance> getAppliances() {
        return appliances;
    }

    public static Main instance() {
        return instance;
    }

    public static Logger logger() {
        return logger;
    }
}
